.tag-unread {
    font-weight: bold;
}

.no-results {
    font-style: italic;
}

img.favicon {
    display: inline-block;
    width: 16px;
    height: 16px;
}

.entry .title {
    display: block;
}

iframe {
    border: 0;
    min-height: 1024px;
}

.entry {
    margin-bottom: 10px;
}
